from ...base import TestBase
from ....attestation.tokentree.token import Token
from ....attestation.tokentree.tree import TokenTree
from ....keyvault.crypto import ECCrypto
from ....keyvault.keys import PublicKey


class TestTree(TestBase):

    def setUp(self) -> None:
        super(TestTree, self).setUp()
        self.crypto = ECCrypto()
        self.private_key = self.crypto.generate_key("curve25519")
        self.public_key = self.private_key.pub()

    def test_create_own_empty(self) -> None:
        """
        Check if we can create our own tree without content.
        """
        tree = TokenTree(private_key=self.private_key)

        self.assertEqual(0, len(tree.elements))
        self.assertEqual(0, len(tree.unchained))
        self.assertEqual(0, len(tree.get_missing()))
        self.assertIsInstance(tree.public_key, PublicKey)

    def test_create_other_empty(self) -> None:
        """
        Check if we can create a tree without content of someone else.
        """
        tree = TokenTree(public_key=self.public_key)

        self.assertEqual(0, len(tree.elements))
        self.assertEqual(0, len(tree.unchained))
        self.assertEqual(0, len(tree.get_missing()))

    def test_own_add(self):
        """
        Check if content is correctly added to the tree.
        """
        tree = TokenTree(private_key=self.private_key)
        content = b"test content"
        token = tree.add(content)

        self.assertEqual(1, len(tree.elements))
        self.assertEqual(0, len(tree.unchained))
        self.assertEqual(0, len(tree.get_missing()))
        self.assertTrue(tree.verify(token))
        self.assertEqual(content, token.content)

    def test_other_add_insequence(self):
        """
        Check if content is correctly added to another's tree, in order.
        """
        tree = TokenTree(public_key=self.public_key)
        real_token = Token(tree.genesis_hash, content=b"some data", private_key=self.private_key)
        pub_token = Token.unserialize(real_token.get_plaintext_signed(), self.public_key)
        token = tree.gather_token(pub_token)

        self.assertEqual(1, len(tree.elements))
        self.assertEqual(0, len(tree.unchained))
        self.assertEqual(0, len(tree.get_missing()))
        self.assertTrue(tree.verify(pub_token))
        self.assertIsNone(token.content)

    def test_other_add_outsequence(self):
        """
        Check if content is not added to another's tree if it is not linked to existing data.
        """
        tree = TokenTree(public_key=self.public_key)
        real_token = Token(b"ab" * 16, content=b"some data", private_key=self.private_key)
        pub_token = Token.unserialize(real_token.get_plaintext_signed(), self.public_key)
        token = tree.gather_token(pub_token)

        self.assertEqual(0, len(tree.elements))
        self.assertEqual(1, len(tree.unchained))
        self.assertListEqual([b"ab" * 16], list(tree.get_missing()))
        self.assertFalse(tree.verify(pub_token))
        self.assertIsNone(token)

    def test_other_add_outsequence_overflow(self):
        """
        Check if content is not added to another's tree if it is not linked to existing data.
        Remove the oldest data when too much data is pending.
        """
        tree = TokenTree(public_key=self.public_key)
        tree.unchained_max_size = 1
        real_token1 = Token(b"ab" * 16, content=b"some data", private_key=self.private_key)
        real_token2 = Token(b"cd" * 16, content=b"some other data", private_key=self.private_key)
        pub_token1 = Token.unserialize(real_token1.get_plaintext_signed(), self.public_key)
        pub_token2 = Token.unserialize(real_token2.get_plaintext_signed(), self.public_key)
        tree.gather_token(pub_token1)
        tree.gather_token(pub_token2)

        self.assertEqual(0, len(tree.elements))
        self.assertEqual(1, len(tree.unchained))
        self.assertListEqual([b"cd" * 16], list(tree.get_missing()))
        self.assertFalse(tree.verify(pub_token1))
        self.assertFalse(tree.verify(pub_token2))

    def test_other_add_duplicate(self):
        """
        Check if tokens are not added twice (by branch position and content hash).
        """
        tree = TokenTree(public_key=self.public_key)
        real_token = Token(tree.genesis_hash, content=b"some data", private_key=self.private_key)
        pub_token = Token.unserialize(real_token.get_plaintext_signed(), self.public_key)
        token = tree.gather_token(pub_token)
        tree.gather_token(pub_token)

        self.assertEqual(1, len(tree.elements))
        self.assertEqual(0, len(tree.unchained))
        self.assertEqual(0, len(tree.get_missing()))
        self.assertTrue(tree.verify(pub_token))
        self.assertIsNone(token.content)

    def test_other_add_duplicate_wcontent(self):
        """
        Check if a token passes on its content to a duplicate entry without content.
        """
        tree = TokenTree(public_key=self.public_key)
        real_token = Token(tree.genesis_hash, content=b"some data", private_key=self.private_key)
        pub_token = Token.unserialize(real_token.get_plaintext_signed(), self.public_key)
        token = tree.gather_token(pub_token)
        tree.gather_token(real_token)

        self.assertEqual(1, len(tree.elements))
        self.assertEqual(0, len(tree.unchained))
        self.assertEqual(0, len(tree.get_missing()))
        self.assertTrue(tree.verify(pub_token))
        self.assertTrue(tree.verify(token))
        self.assertEqual(b"some data", token.content)

    def test_serialize_public(self):
        """
        Check if the public tree data is serialized correctly.
        """
        tree = TokenTree(private_key=self.private_key)
        tree.add(b"data1")
        tree.add(b"data2")
        tree.add(b"data3")

        self.assertEqual(128 * 3, len(tree.serialize_public()))

    def test_serialize_public_partial(self):
        """
        Check if the public tree data is partially serialized correctly.
        """
        tree = TokenTree(private_key=self.private_key)
        first_token = tree.add(b"data1")
        second_token = tree.add(b"data2", first_token)
        tree.add(b"data3", second_token)

        self.assertEqual(128 * 2, len(tree.serialize_public(second_token)))

    def test_unserialize_public(self):
        """
        Check if serialized public tree data can be paired with its source data.
        """
        public_key = self.crypto.key_from_public_bin(b'LibNaCLPK:/N\xc5\xd1#\xd4\xc5\x02\xca\xb4\xa4\xd4vKD\xf1"\xf0'
                                                     b'1,\\\xde\x14\x87\xa9\xf6T\x90\xd9\xb0qk\xdbPS\xfbqm\xc1,i\xca'
                                                     b'\x88\x7fm\xe8\\\x0f\xe9\xee\xec\xce\xbeN\xdc\x94\xc4\x84\'\x8b'
                                                     b'\xb8\x8e\x1b\xc4')
        s = (b'\xb1\xa6\xf8#\x91q\xa2\x95 \xb7\xa1\x0c"6 1&\xdd\xf6\xf53\x0b\xe6\x17\xe3 H\xa6\xe0\xfa\x1d4\x1f\xef'
             b'\xf3:1\xfe{\xd7\x10\xf78\xec\x96\xe9Z\x8a\x85Q\xa2\x04z\xbdF0\x9b{%\x94\xe6\x065\x95(\xc9\x9c(\x87\xe0'
             b'Z\xff\xcf\xfd\x880}\\\xa0\xf5WD\x00\x14D.\x8d\x99p\xc6%3:\xc1\x1d\xaeP\x8bU\xc5\xe6\xe2,\xc01MN\x96\xdf'
             b'\xad=\xf1J\x1f\x12\x19B\x96e\x98EV\xa0-i\xb1\xa3\x04\xdd\xc2\xef\\\xc7\x93\xb2?0\xba4\xda\xac\x87\xd0o'
             b'\xe0cl\xd4\xe5\x86\xdb\xb1K\x84\xe3oE\x1e\xcb\x9b\xf3\xab>\xc9<\xbe\xcf*\xe8(\x0f\rx!\x11\xab2\x02\xdf'
             b'\xa6\xee\xf6yr\x03\xb8\x86\xe2\x15\xfc{n@\x9a\xcb\xb8I\xfa\x91\xf6\xb5\xac\x82\x9e\x18;\xcd\xba\x13\x94'
             b'\x91{%,[k\xe9\xb1,E\xb1\n$\xb5\xd2\xfe\xeb\xe2\x1bS4;0\xc2\xd2\xf5*\x98\xacxc\xebb\x9d\'<E\x8az\xf6\xc9'
             b'Z)x\x83\x02\xc2\x87\x92KU`\x80\x9btH\x17\xd00\xea\x8e4\x19\xc7\x1c5\xff\xdb\xc4\xd5e\xdfw\x02\xb0N\xe9'
             b'\x82Lu\xb3\x10\xb7i\xb0\xa0\xc8R\x1a@\x96\x8a\xf4=\xafb\xce\xc5\xab\x1ae\xdce\x80\x93\x19  \x89|\xf2'
             b'\x0cM^q\xb7\xa8{\xa9\xf2Y\x8d<\xb5\x8e\xec6S_\x96\x03\x9c\xd1\xb1\x03\x9b\xe7\x89{_\xd8`\x91\xe8{>\x07'
             b'M\xce\xd9!\x88]\xe6\x1c\xf7\x99\x9f\xc3au\xdd\x9f\\\xaf0\xe0L\xfcu\xd0e\x84\x05')
        tree = TokenTree(public_key=public_key)
        tree.unserialize_public(s)

        self.assertEqual(3, len(tree.elements))

        for token in tree.elements.values():
            self.assertTrue(tree.verify(token))

        # By design, the TokenTree does not keep the mapping between content and Tokens.
        # But we can show the three different values are captured by the three Tokens.
        for content in [b"data1", b"data2", b"data3"]:
            self.assertTrue(any([token.receive_content(content) for token in tree.elements.values()]))



import os
import sys
import time
from asyncio import ensure_future, get_event_loop, sleep
from os import chdir, getcwd, mkdir, path
from random import randint

# Check if we are running from the root directory
# If not, modify our path so that we can import IPv8
try:
    import ipv8
    del ipv8
except ImportError:
    import __scriptpath__  # noqa: F401


from ipv8.configuration import get_default_configuration  # pylint: disable=ungrouped-imports

from ipv8_service import IPv8, _COMMUNITIES, _WALKERS  # pylint: disable=ungrouped-imports


START_TIME = time.time()
RESULTS = {}


# Wait until we get a non-tracker response
# Once all overlays have finished, stop the script
def custom_intro_response_cb(self, peer, dist, payload):
    if (peer.address not in self.network.blacklist) and (self.__class__.__name__ not in RESULTS):
        RESULTS[self.__class__.__name__] = time.time() - START_TIME
        print(self.__class__.__name__, "found a peer!", file=sys.stderr)
        if len(get_default_configuration()['overlays']) == len(RESULTS):
            get_event_loop().stop()


# If it takes longer than 30 seconds to find anything, abort the experiment and set the intro time to -1.0
async def on_timeout():
    await sleep(30)
    for definition in get_default_configuration()['overlays']:
        if definition['class'] not in RESULTS:
            RESULTS[definition['class']] = -1.0
            print(definition['class'], "found no peers at all!", file=sys.stderr)
    get_event_loop().stop()


async def start_communities():
    # Override the Community master peers so we don't interfere with the live network
    # Also hook in our custom logic for introduction responses
    for community_cls in _COMMUNITIES.values():
        community_cls.community_id = os.urandom(20)
        community_cls.introduction_response_callback = custom_intro_response_cb

    # Create two peers with separate working directories
    previous_workdir = getcwd()
    for i in [1, 2]:
        configuration = get_default_configuration()
        configuration['port'] = 12000 + randint(0, 10000)
        configuration['logger']['level'] = "CRITICAL"
        for overlay in configuration['overlays']:
            overlay['walkers'] = [walker for walker in overlay['walkers'] if walker['strategy'] in _WALKERS]
        workdir = path.abspath(path.join(path.dirname(__file__), str(i)))
        if not path.exists(workdir):
            mkdir(workdir)
        chdir(workdir)
        await IPv8(configuration).start()
        chdir(previous_workdir)

# Actually start running everything, this blocks until the experiment finishes
ensure_future(on_timeout())
ensure_future(start_communities())
get_event_loop().run_forever()

# Print the introduction times for all default Communities, sorted alphabetically.
print(','.join(['%.4f' % RESULTS[key] for key in sorted(RESULTS)]))



import json

from ...base import TestBase
from ....attestation.identity.database import IdentityDatabase
from ....attestation.identity.manager import IdentityManager
from ....keyvault.crypto import ECCrypto


class TestIdentityManager(TestBase):

    def setUp(self) -> None:
        super(TestIdentityManager, self).setUp()
        self.crypto = ECCrypto()
        self.private_key = self.crypto.generate_key("curve25519")
        self.public_key = self.private_key.pub()
        self.authority_private_key = self.crypto.generate_key("curve25519")
        self.authority_public_key = self.authority_private_key.pub()
        self.manager = IdentityManager()

    def forget_identities(self):
        self.manager.pseudonyms.clear()
        self.manager.database = IdentityDatabase(":memory:")
        self.manager.database.open()

    def test_create_identity(self):
        """
        Test if a new pseudonym is correctly created.
        """
        pseudonym = self.manager.get_pseudonym(self.private_key)

        self.assertEqual([], pseudonym.get_credentials())

    def test_substantiate_empty(self):
        """
        Check if an empty identity disclosure is loaded and valid.
        """
        valid, pseudonym = self.manager.substantiate(self.public_key, b'', b'', b'', b'')

        self.assertTrue(valid)
        self.assertEqual([], pseudonym.get_credentials())

    def test_create_credential(self):
        """
        Test creating a credential without attestations.
        """
        pseudonym = self.manager.get_pseudonym(self.private_key)
        pseudonym.create_credential(b'ab' * 16, {'some_key': 'some_value'})

        self.assertEqual(1, len(pseudonym.get_credentials()))
        self.assertEqual(0, len(pseudonym.get_credentials()[0].attestations))
        self.assertDictEqual({'some_key': 'some_value'},
                             json.loads(pseudonym.get_credentials()[0].metadata.serialized_json_dict))

    def test_substantiate_credential_update(self):
        """
        Test substantiating a credential without attestations, with existing metadata.
        """
        pseudonym = self.manager.get_pseudonym(self.private_key)
        pseudonym.create_credential(b'ab' * 16, {'some_key': 'some_value'})
        metadata, tokens, attestations, authorities = pseudonym.disclose_credentials(pseudonym.get_credentials(), set())

        self.manager.pseudonyms.clear()

        valid, public_pseudonym = self.manager.substantiate(pseudonym.public_key, metadata, tokens, attestations,
                                                            authorities)

        self.assertTrue(valid)
        self.assertEqual(1, len(public_pseudonym.get_credentials()))
        self.assertEqual(0, len(public_pseudonym.get_credentials()[0].attestations))
        self.assertDictEqual({'some_key': 'some_value'},
                             json.loads(public_pseudonym.get_credentials()[0].metadata.serialized_json_dict))

    def test_substantiate_credential_no_metadata(self):
        """
        Test substantiating a credential without attestations, without metadata.

        This situation is a bit tricky:
         - The path to the root from the one disclosed Token is valid.
         - No Metadata is known for this Token and it therefore does not form a credential.
        """
        pseudonym = self.manager.get_pseudonym(self.private_key)
        pseudonym.create_credential(b'ab' * 16, {'some_key': 'some_value'})
        metadata, tokens, attestations, authorities = pseudonym.disclose_credentials(pseudonym.get_credentials(), set())

        self.forget_identities()

        valid, public_pseudonym = self.manager.substantiate(pseudonym.public_key, b'', tokens, attestations,
                                                            authorities)

        self.assertTrue(valid)
        self.assertEqual(0, len(public_pseudonym.get_credentials()))
        self.assertEqual(1, len(public_pseudonym.tree.elements))

    def test_substantiate_credential_with_metadata(self):
        """
        Test substantiating a credential without attestations.
        """
        pseudonym = self.manager.get_pseudonym(self.private_key)
        pseudonym.create_credential(b'ab' * 16, {'some_key': 'some_value'})
        metadata, tokens, attestations, authorities = pseudonym.disclose_credentials(pseudonym.get_credentials(), set())

        self.forget_identities()

        valid, public_pseudonym = self.manager.substantiate(pseudonym.public_key, metadata, tokens, attestations,
                                                            authorities)

        self.assertTrue(valid)
        self.assertEqual(1, len(public_pseudonym.get_credentials()))
        self.assertEqual(0, len(public_pseudonym.get_credentials()[0].attestations))
        self.assertDictEqual({'some_key': 'some_value'},
                             json.loads(public_pseudonym.get_credentials()[0].metadata.serialized_json_dict))

    def test_substantiate_credential_full(self):
        """
        Test substantiating a typical credential.
        """
        pseudonym = self.manager.get_pseudonym(self.private_key)
        pseudonym.create_credential(b'ab' * 16, {'some_key': 'some_value'})

        attestation = pseudonym.create_attestation(pseudonym.get_credentials()[0].metadata, self.authority_private_key)
        pseudonym.add_attestation(self.authority_public_key, attestation)

        metadata, tokens, attestations, authorities = pseudonym.disclose_credentials(pseudonym.get_credentials(),
                                                                                     {attestation.get_hash()})

        self.forget_identities()

        valid, public_pseudonym = self.manager.substantiate(pseudonym.public_key, metadata, tokens, attestations,
                                                            authorities)

        self.assertTrue(valid)
        self.assertEqual(1, len(public_pseudonym.get_credentials()))
        self.assertEqual(1, len(public_pseudonym.get_credentials()[0].attestations))
        self.assertDictEqual({'some_key': 'some_value'},
                             json.loads(public_pseudonym.get_credentials()[0].metadata.serialized_json_dict))

    def test_substantiate_credential_partial(self):
        """
        Test substantiating a typical credential, with partial disclosure.
        """
        pseudonym = self.manager.get_pseudonym(self.private_key)
        pseudonym.create_credential(b'ab' * 16, {'some_key': 'some_value'})

        attestation = pseudonym.create_attestation(pseudonym.get_credentials()[0].metadata, self.authority_private_key)
        pseudonym.add_attestation(self.authority_public_key, attestation)

        attestation2 = pseudonym.create_attestation(pseudonym.get_credentials()[0].metadata, self.private_key)
        pseudonym.add_attestation(self.public_key, attestation2)

        metadata, tokens, attestations, authorities = pseudonym.disclose_credentials(pseudonym.get_credentials(),
                                                                                     {attestation.get_hash()})

        self.forget_identities()

        valid, public_pseudonym = self.manager.substantiate(pseudonym.public_key, metadata, tokens, attestations,
                                                            authorities)

        self.assertTrue(valid)
        self.assertEqual(1, len(public_pseudonym.get_credentials()))
        self.assertEqual(1, len(public_pseudonym.get_credentials()[0].attestations))
        self.assertDictEqual({'some_key': 'some_value'},
                             json.loads(public_pseudonym.get_credentials()[0].metadata.serialized_json_dict))



import json
import os
import typing
from binascii import hexlify, unhexlify
from time import time

from .attestation import Attestation
from .manager import IdentityManager
from .payload import AttestPayload, DiclosePayload, MissingResponsePayload, RequestMissingPayload
from ...bootstrapping.dispersy.bootstrapper import DispersyBootstrapper
from ...community import Community, DEFAULT_MAX_PEERS
from ...configuration import DISPERSY_BOOTSTRAPPER
from ...lazy_community import lazy_wrapper
from ...peer import Peer
from ...peerdiscovery.discovery import RandomWalk
from ...peerdiscovery.network import Network
from ...types import Credential, Endpoint, Metadata, PrivateKey, PseudonymManager, PublicKey, Token

SAFE_UDP_PACKET_LENGTH = 1296


class IdentityCommunity(Community):

    community_id = unhexlify('d5889074c1e4c50423cdb6e9307ee0ca5695ead7')

    def __init__(self, my_peer: Peer, endpoint: Endpoint, network: Network = None, max_peers: int = DEFAULT_MAX_PEERS,
                 anonymize: bool = True, identity_manager: IdentityManager = None, working_directory: str = "."):
        if network is None:
            network = Network()
        super(IdentityCommunity, self).__init__(my_peer, endpoint, network, max_peers, anonymize)
        if identity_manager is None:
            dbpath = ":memory:" if working_directory == ":memory:" else os.path.join(working_directory, "sqlite",
                                                                                     "identity.db")
            identity_manager = IdentityManager(database_path=dbpath)

        # Dict of hash -> (attribute_name, date, public_key)
        self.known_attestation_hashes: typing.Dict[bytes, typing.Tuple[str, float, PublicKey,
                                                                       typing.Optional[typing.Dict[str, str]]]] = {}

        self.identity_manager = identity_manager
        self.pseudonym_manager = identity_manager.get_pseudonym(self.my_peer.key)

        # We assume other people try to cheat us with trees.
        # We don't attack ourselves though and just maintain a chain of attributes per pseudonym.
        self.token_chain: typing.List[Token] = []
        self.metadata_chain = []
        self.attestation_chain = []
        self.permissions: typing.Dict[Peer, int] = {}  # Map of peer to highest index

        # Pick the longest chain in case of bugs or malicious behavior.. hello Bitcoin.
        for token in self.pseudonym_manager.tree.elements.values():
            chain = self.pseudonym_manager.tree.get_root_path(token)
            if len(chain) > len(self.token_chain):
                self.token_chain = chain
        for credential in self.pseudonym_manager.get_credentials():
            for token in self.token_chain:
                if credential.metadata.token_pointer == token.get_hash():
                    self.metadata_chain.append(credential.metadata)
                    self.attestation_chain.append(credential.attestations)
                    break

        # Register messages
        self.add_message_handler(DiclosePayload, self.on_disclosure)
        self.add_message_handler(AttestPayload, self.on_attest)
        self.add_message_handler(RequestMissingPayload, self.on_request_missing)
        self.add_message_handler(MissingResponsePayload, self.on_missing_response)

    def pad_hash(self, attribute_hash: bytes) -> bytes:
        """
        Pad an old-style SHA-1 hash into the new 32 byte SHA3-256 space.
        """
        self.logger.debug("Padding deprecated SHA-1 hash to 32 bytes, use SHA3-256 instead!")
        return b'SHA-1\x00\x00\x00\x00\x00\x00\x00' + attribute_hash

    def add_known_hash(self,
                       attribute_hash: bytes,
                       name: str,
                       public_key: PublicKey,
                       metadata: typing.Optional[typing.Dict[str, str]] = None) -> None:
        """
        We know about this hash+peer combination. Thus we can handle sign requests for it.
        """
        if len(attribute_hash) == 20:
            attribute_hash = self.pad_hash(attribute_hash)
        self.known_attestation_hashes[attribute_hash] = (name, time(), public_key, metadata)

    def get_attestation_by_hash(self, attribute_hash: bytes) -> typing.Optional[Metadata]:
        """
        Get the Metadata object for a particular attribute hash, if it exists.
        """
        if len(attribute_hash) == 20:
            attribute_hash = self.pad_hash(attribute_hash)
        for credential in self.pseudonym_manager.get_credentials():
            token = self.pseudonym_manager.tree.elements[credential.metadata.token_pointer]
            if token.content_hash == attribute_hash:
                return credential.metadata
        return None

    def should_sign(self,
                    pseudonym: PseudonymManager,
                    metadata: Metadata) -> bool:
        """
        Has the user asked us to sign for some metadata and is it still valid?
        """
        transaction = json.loads(metadata.serialized_json_dict)
        requested_keys = set(transaction.keys())
        if metadata.token_pointer not in pseudonym.tree.elements:
            return False
        attribute_hash = pseudonym.tree.elements[metadata.token_pointer].content_hash
        if "name" not in requested_keys or "date" not in requested_keys or "schema" not in requested_keys:
            self.logger.debug(f"Not signing {str(metadata)}, it doesn't include the required fields!")
            return False
        if attribute_hash not in self.known_attestation_hashes:
            self.logger.debug(f"Not signing {str(metadata)}, it doesn't point to known content!")
            return False
        if pseudonym.public_key.key_to_bin() != self.known_attestation_hashes[attribute_hash][2]:
            self.logger.debug(f"Not signing {str(metadata)}, attribute doesn't belong to key!")
            return False
        # Refuse to sign blocks older than 5 minutes
        if time() > self.known_attestation_hashes[attribute_hash][1] + 300:
            self.logger.debug(f"Not signing {str(metadata)}, timed out!")
            return False
        if transaction['name'] != self.known_attestation_hashes[attribute_hash][0]:
            self.logger.debug(f"Not signing {str(metadata)}, name does not match!")
            return False
        if (self.known_attestation_hashes[attribute_hash][3] is not None
                and ({k: v for k, v in transaction.items() if k not in ["name", "date", "schema"]}
                     != self.known_attestation_hashes[attribute_hash][3])):
            self.logger.debug(f"Not signing {str(metadata)}, metadata does not match!")
            return False
        for attestation in pseudonym.database.get_attestations_over(metadata):
            if any(authority == self.my_peer.public_key.key_to_bin()
                   for authority in pseudonym.database.get_authority(attestation)):
                self.logger.debug(f"Not signing {str(metadata)}, already attested!")
                return False
        return True

    def _fit_disclosure(self, disclosure: typing.Tuple[bytes, bytes, bytes, bytes])\
            -> typing.Tuple[bytes, bytes, bytes, bytes]:
        """
        Fit a disclosure (metadata, tokens, attestations and authorities) to a UDP packet.

        This comes down to stripping tokens until the serialization fits in a UDP packet, as tokens can be shown
        to be missing from a disclosure and will be retrieved on demand
        """
        token_size = 64 + self.crypto.get_signature_length(self.my_peer.key)
        metadata, tokens, attestations, authorities = disclosure
        meta_len = len(metadata) + len(attestations) + len(authorities)
        if meta_len + len(tokens) > SAFE_UDP_PACKET_LENGTH:
            packet_space = SAFE_UDP_PACKET_LENGTH - meta_len
            if packet_space < 0:
                self.logger.warning(f"Attempting to disclose with packet of length {meta_len}, hoping for the best!")
            packet_space = max(0, packet_space)
            trim_len = packet_space // token_size
            tokens = tokens[-trim_len * token_size:]
        return metadata, tokens, attestations, authorities

    def _received_disclosure_for_attest(self,
                                        peer: Peer,
                                        disclosure: typing.Tuple[bytes, bytes, bytes, bytes]) -> None:
        """
        Attempt to insert a disclosure into our database and request more if we are still missing tokens.
        """
        solicited = any(t[2] == peer.public_key.key_to_bin() for t in self.known_attestation_hashes.values())
        if solicited:
            correct, pseudonym = self.identity_manager.substantiate(peer.public_key, *disclosure)
            required_attributes = [attribute_hash for attribute_hash in self.known_attestation_hashes
                                   if self.known_attestation_hashes[attribute_hash][2] == peer.public_key.key_to_bin()]
            known_attributes = [token.content_hash for token in pseudonym.tree.elements.values()]
            if correct and any(attribute_hash in known_attributes for attribute_hash in required_attributes):
                for credential in pseudonym.get_credentials():
                    if self.should_sign(pseudonym, credential.metadata):
                        self.logger.info(f"Attesting to {str(credential.metadata)}")
                        attestation = pseudonym.create_attestation(credential.metadata, self.my_peer.key)
                        pseudonym.add_attestation(self.my_peer.public_key, attestation)
                        self.ez_send(peer, AttestPayload(attestation.get_plaintext_signed()))
            for attribute_hash in required_attributes:
                if attribute_hash not in known_attributes:
                    self.logger.info("Missing information for attestation %s, requesting more!",
                                     hexlify(attribute_hash).decode())
                    self.ez_send(peer, RequestMissingPayload(len(pseudonym.tree.elements)))
        else:
            self.logger.warning("Received unsolicited disclosure from %s, dropping!", str(peer))

    def request_attestation_advertisement(self,
                                          peer: Peer,
                                          attribute_hash: bytes,
                                          name: str,
                                          block_type: str = "id_metadata",
                                          metadata: typing.Optional[dict] = None) -> None:
        """
        Request a peer to sign for our attestation advertisement.
        :param peer: the attestor of our block
        :param attribute_hash: the hash of the attestation
        :param name: the name of the attribute (metadata)
        :param block_type: the type of block (from identity_foromats.py)
        :param metadata: custom additional metadata
        """
        credential = self.self_advertise(attribute_hash, name, block_type, metadata)
        self.permissions[peer] = len(self.token_chain)
        disclosure = self.pseudonym_manager.disclose_credentials([credential], set())
        self.ez_send(peer, DiclosePayload(*self._fit_disclosure(disclosure)))

    def self_advertise(self,
                       attribute_hash: bytes,
                       name: str,
                       block_type: str = "id_metadata",
                       metadata: typing.Optional[dict] = None) -> Credential:
        """
        Self-sign an attribute.

        :param attribute_hash: he hash of the attestation
        :param name: the name of the attribute (metadata)
        :param block_type: the type of block (from identity_formats.py)
        :param metadata: custom additional metadata
        """
        if len(attribute_hash) == 20:
            attribute_hash = self.pad_hash(attribute_hash)

        # Construct metadata fields
        extended_metadata = {
            "name": name,
            "schema": block_type,
            "date": time()
        }
        if metadata:
            extended_metadata.update(metadata)

        # Create credential
        credential = self.pseudonym_manager.create_credential(attribute_hash, extended_metadata,
                                                              self.metadata_chain[-1] if self.metadata_chain else None)

        # Construct chain data view
        self.attestation_chain.append(credential.attestations)
        self.metadata_chain.append(credential.metadata)
        self.token_chain.append(self.pseudonym_manager.tree.elements[credential.metadata.token_pointer])

        return credential

    @lazy_wrapper(DiclosePayload)
    def on_disclosure(self, peer: Peer, disclosure: DiclosePayload) -> None:
        """
        Someone disclosed their attributes to us.
        Attempt to insert them into our database and check if we are still missing some.
        """
        self._received_disclosure_for_attest(peer, (disclosure.metadata, disclosure.tokens, disclosure.attestations,
                                                    disclosure.authorities))

    @lazy_wrapper(AttestPayload)
    def on_attest(self, peer: Peer, payload: AttestPayload) -> None:
        """
        Someone made an attestation for us, try to insert it into our database.
        """
        attestation = Attestation.unserialize(payload.attestation, peer.public_key)
        if self.pseudonym_manager.add_attestation(peer.public_key, attestation):
            self.logger.info(f"Received attestation from {str(peer)}!")
        else:
            self.logger.warning(f"Received invalid attestation from {str(peer)}!")

    @lazy_wrapper(RequestMissingPayload)
    def on_request_missing(self, peer: Peer, request: RequestMissingPayload) -> None:
        """
        Someone requested tokens from us. If they are permitted to see them, send them over.

        Note that tokens do not have indices publicly, so instead of sending missing tokens one at a time,
        we send a range, starting from the lowest index that is missing.
        """
        out = b''
        permitted = self.token_chain[:self.permissions.get(peer, 0)]
        for index, token in enumerate(permitted):
            if index >= request.known:
                serialized = token.get_plaintext_signed()
                if len(out) + len(serialized) > SAFE_UDP_PACKET_LENGTH:
                    break
                out += serialized
        self.ez_send(peer, MissingResponsePayload(out))

    @lazy_wrapper(MissingResponsePayload)
    def on_missing_response(self, peer: Peer, response: MissingResponsePayload) -> None:
        """
        We received tokens, attempt to insert them into our database and check if we are still missing some.
        """
        self._received_disclosure_for_attest(peer, (b'', response.tokens, b'', b''))


async def create_community(private_key: PrivateKey, ipv8, identity_manager: IdentityManager,
                           endpoint: typing.Optional[Endpoint] = None, working_directory: typing.Optional[str] = None,
                           anonymize: bool = True,
                           rendezvous_token: typing.Optional[bytes] = None) -> IdentityCommunity:
    my_peer = Peer(private_key)
    if endpoint is None:
        endpoint = await ipv8.produce_anonymized_endpoint()
    working_directory_str: str = (ipv8.configuration.get("working_directory") if not working_directory
                                  else working_directory)
    overlay_cls = IdentityCommunity
    if rendezvous_token is not None:
        token_str = hexlify(rendezvous_token).decode()
        rendezvous_id = bytes(b ^ rendezvous_token[i] if i < len(rendezvous_token) else b
                              for i, b in enumerate(IdentityCommunity.community_id))
        overlay_cls = type(f"IdentityCommunity-{token_str}", (IdentityCommunity, ), {  # type:ignore
            'community_id': rendezvous_id
        })
    community = overlay_cls(my_peer, endpoint, identity_manager=identity_manager,
                            working_directory=working_directory_str, anonymize=anonymize)
    community.bootstrappers = [DispersyBootstrapper(**DISPERSY_BOOTSTRAPPER['init'])]
    ipv8.add_strategy(community, RandomWalk(community), -1)
    return community



import logging
import socket
import sys
import time
from asyncio import CancelledError, DatagramProtocol, Future, ensure_future, gather, get_event_loop
from binascii import hexlify
from collections import deque
from struct import unpack_from
from traceback import format_exception

from ...taskmanager import TaskManager
from ...util import succeed

ORIGINATOR = 0
EXIT_NODE = 1
ORIGINATOR_SALT = 2
EXIT_NODE_SALT = 3
ORIGINATOR_SALT_EXPLICIT = 4
EXIT_NODE_SALT_EXPLICIT = 5

PEER_SOURCE_UNKNOWN = 0
PEER_SOURCE_DHT = 1
PEER_SOURCE_PEX = 2

PEER_FLAG_RELAY = 1
PEER_FLAG_EXIT_BT = 2
PEER_FLAG_EXIT_IPV8 = 4
PEER_FLAG_SPEED_TEST = 8

# Data circuits are general purpose circuits for exiting data
CIRCUIT_TYPE_DATA = 'DATA'

# The other circuits are supposed to end in a connectable node, not allowed to exit
# anything else than IPv8 messages, used for setting up end-to-end circuits
CIRCUIT_TYPE_IP_SEEDER = 'IP_SEEDER'
CIRCUIT_TYPE_RP_SEEDER = 'RP_SEEDER'
CIRCUIT_TYPE_RP_DOWNLOADER = 'RP_DOWNLOADER'

CIRCUIT_STATE_READY = 'READY'
CIRCUIT_STATE_EXTENDING = 'EXTENDING'
CIRCUIT_STATE_CLOSING = 'CLOSING'

CIRCUIT_ID_PORT = 1024
PING_INTERVAL = 7.5

# Reasons for sending destroy messages. Code 0 must not be used for legacy reasons.
DESTROY_REASON_UNKNOWN = 1
DESTROY_REASON_SHUTDOWN = 2
DESTROY_REASON_FORWARD = 3
DESTROY_REASON_UNNEEDED = 4
DESTROY_REASON_LEAVE_SWARM = 5
DESTROY_REASON_FORBIDDEN = 6


class DataChecker(object):

    @staticmethod
    def could_be_utp(data):
        if len(data) < 20:
            return False
        byte1, byte2 = unpack_from('!BB', data)
        # Type should be 0..4, Ver should be 1
        if not (0 <= (byte1 >> 4) <= 4 and (byte1 & 15) == 1):
            return False
        # Extension should be 0..2
        if not (0 <= byte2 <= 2):
            return False
        return True

    @staticmethod
    def could_be_udp_tracker(data):
        # For the UDP tracker protocol the action field is either at position 0 or 8, and should be 0..3
        if len(data) >= 8 and (0 <= unpack_from('!I', data, 0)[0] <= 3)\
                or len(data) >= 12 and (0 <= unpack_from('!I', data, 8)[0] <= 3):
            return True
        return False

    @staticmethod
    def could_be_dht(data):
        try:
            if len(data) > 1 and data[0:1] == b'd' and data[-1:] == b'e':
                return True
        except TypeError:
            pass
        return False

    @staticmethod
    def could_be_bt(data):
        return (DataChecker.could_be_utp(data)
                or DataChecker.could_be_udp_tracker(data)
                or DataChecker.could_be_dht(data))

    @staticmethod
    def could_be_ipv8(data):
        return len(data) >= 23 and data[0:1] == b'\x00' and data[1:2] in [b'\x01', b'\x02']


class Tunnel(object):

    def __init__(self, circuit_id, peer):
        self.circuit_id = circuit_id
        self._peer = peer
        self.creation_time = time.time()
        self.last_activity = time.time()
        self.bytes_up = self.bytes_down = 0
        self.logger = logging.getLogger(self.__class__.__name__)

    @property
    def peer(self):
        return self._peer

    def beat_heart(self):
        self.last_activity = time.time()


class TunnelExitSocket(Tunnel, DatagramProtocol, TaskManager):

    def __init__(self, circuit_id, peer, overlay):
        Tunnel.__init__(self, circuit_id, peer)
        TaskManager.__init__(self)
        self.overlay = overlay
        self.transport = None
        self.queue = deque(maxlen=10)
        self.enabled = False

    def enable(self):
        if not self.enabled:
            self.enabled = True

            async def create_transport():
                self.transport, _ = await get_event_loop().create_datagram_endpoint(lambda: self,
                                                                                    local_addr=('0.0.0.0', 0))
                # Send any packets that have been waiting while the transport was being created
                while self.queue:
                    self.sendto(*self.queue.popleft())
            self.register_task("create_transport", create_transport)

    def sendto(self, data, destination):
        if not self.transport:
            self.queue.append((data, destination))
            return

        self.beat_heart()
        if self.is_allowed(data):
            def on_ip_address(future):
                try:
                    ip_address = future.result()
                except (CancelledError, Exception) as e:
                    self.logger.error("Can't resolve ip address for %s. Failure: %s", destination[0], e)
                    return

                self.logger.debug("Resolved hostname %s to ip_address %s", destination[0], ip_address)
                try:
                    self.transport.sendto(data, (ip_address, destination[1]))
                    self.bytes_up += len(data)
                except socket.error as e:
                    self.logger.error("Failed to write to transport. Destination: %r error: %r", destination, e)

            try:
                socket.inet_aton(destination[0])
                on_ip_address(succeed(destination[0]))
            except (OSError, ValueError):
                task = ensure_future(self.resolve(destination[0]))
                # If this also fails, the TaskManager logs the packet.
                # The host probably really does not exist.
                self.register_anonymous_task("resolving_%r" % destination[0], task,
                                             ignore=(OSError, ValueError)).add_done_callback(on_ip_address)

    async def resolve(self, host):
        # Using asyncio's getaddrinfo since the aiodns resolver seems to have issues.
        # Returns [(family, type, proto, canonname, sockaddr)].
        infos = await get_event_loop().getaddrinfo(host, 0, family=socket.AF_INET)
        return infos[0][-1][0]

    def datagram_received(self, data, source):
        self.beat_heart()
        self.bytes_down += len(data)
        if self.is_allowed(data):
            try:
                self.tunnel_data(source, data)
            except Exception:
                self.logger.error("Exception occurred while handling incoming exit node data!\n"
                                  + ''.join(format_exception(*sys.exc_info())))
        else:
            self.logger.warning("Dropping forbidden packets to exit socket with circuit_id %d", self.circuit_id)

    def is_allowed(self, data):
        is_bt = DataChecker.could_be_bt(data)
        is_ipv8 = DataChecker.could_be_ipv8(data)

        if not (is_bt and PEER_FLAG_EXIT_BT in self.overlay.settings.peer_flags) \
           and not (is_ipv8 and PEER_FLAG_EXIT_IPV8 in self.overlay.settings.peer_flags) \
           and not (is_ipv8 and self.overlay._prefix == data[:22]):
            self.logger.error("Dropping data packets, refusing to be an exit node (BT=%s, IPv8=%s)", is_bt, is_ipv8)
            return False
        return True

    def tunnel_data(self, source, data):
        self.logger.debug("Tunnel data to origin %s for circuit %s", ('0.0.0.0', 0), self.circuit_id)
        self.overlay.send_data(self.peer, self.circuit_id, ('0.0.0.0', 0), source, data)

    async def close(self):
        """
        Closes the UDP socket if enabled and cancels all pending deferreds.
        :return: A deferred that fires once the UDP socket has closed.
        """
        # The resolution tasks can't be cancelled, so we need to wait for
        # them to finish.
        await self.shutdown_task_manager()
        if self.transport:
            self.transport.close()
            self.transport = None


class Circuit(Tunnel):

    def __init__(self, circuit_id, goal_hops=0, ctype=CIRCUIT_TYPE_DATA, required_exit=None, info_hash=None):
        super(Circuit, self).__init__(circuit_id, None)
        self.goal_hops = goal_hops
        self.ctype = ctype
        self.required_exit = required_exit
        self.info_hash = info_hash

        self.ready = Future()
        self.closing_info = ''
        self._closing = False
        self._hops = []
        self.unverified_hop = None
        self.hs_session_keys = None
        self.e2e = False
        self.relay_early_count = 0

    @property
    def peer(self):
        if self._hops:
            return self._hops[0].peer
        elif self.unverified_hop:
            return self.unverified_hop.peer

    @property
    def exit_flags(self):
        return self.hops[-1].flags if self.hops else []

    @property
    def hops(self):
        """
        Return a read only tuple version of the hop-list of this circuit
        @rtype tuple[Hop]
        """
        return tuple(self._hops)

    def add_hop(self, hop):
        """
        Adds a hop to the circuits hop collection
        @param Hop hop: the hop to add
        """
        self._hops.append(hop)
        if self.state == CIRCUIT_STATE_READY:
            self.ready.set_result(self)

    @property
    def state(self):
        """
        The circuit state, can be either:
        CIRCUIT_STATE_CLOSING, CIRCUIT_STATE_EXTENDING or CIRCUIT_STATE_READY
        @rtype: str
        """
        if self._closing:
            return CIRCUIT_STATE_CLOSING

        if len(self.hops) < self.goal_hops:
            return CIRCUIT_STATE_EXTENDING
        else:
            return CIRCUIT_STATE_READY

    def close(self, closing_info=''):
        """
        Sets the state of the circuit to CIRCUIT_STATE_CLOSING. This ensures that this circuit
        will not be used to contact new peers.
        """
        self.closing_info = closing_info
        self._closing = True
        if not self.ready.done():
            self.ready.set_result(None)

    def __eq__(self, other):
        return other and self.circuit_id == other.circuit_id


class Hop(object):

    """
    Circuit Hop containing the address, its public key and the first part of
    the Diffie-Hellman handshake
    """

    def __init__(self, peer, flags=None):
        """
        @param Peer peer: IPv8 peer object of the hop
        """
        self.peer = peer
        self.session_keys = None
        self.dh_first_part = None
        self.dh_secret = None
        self.flags = flags or []

    @property
    def public_key(self):
        return self.peer.public_key

    @property
    def node_public_key(self):
        """
        The hop's public_key
        """
        return self.peer.public_key.key_to_bin()

    @property
    def mid(self):
        return self.peer.public_key.key_to_hash()


class RelayRoute(Tunnel):

    """
    Relay object containing the destination circuit, socket address and whether
    it is online or not
    """

    def __init__(self, circuit_id, peer, rendezvous_relay=False):
        """
        @type sock_addr: (str, int)
        @type circuit_id: int
        @return:
        """
        super(RelayRoute, self).__init__(circuit_id, peer)
        self.rendezvous_relay = rendezvous_relay
        # Since the creation of a RelayRoute object is triggered by an extend (which was wrapped in a cell
        # that had the early_relay flag set) we start the count at 1.
        self.relay_early_count = 1


class RendezvousPoint(object):

    def __init__(self, circuit, cookie):
        self.circuit = circuit
        self.cookie = cookie
        self.address = None
        self.ready = Future()


class IntroductionPoint(object):

    def __init__(self, peer, seeder_pk, source=PEER_SOURCE_UNKNOWN, last_seen=None):
        self.peer = peer
        self.seeder_pk = seeder_pk
        self.source = source
        self.last_seen = int(time.time()) if last_seen is None else last_seen

    def __eq__(self, other):
        return self.peer == other.peer and self.seeder_pk == other.seeder_pk

    def __hash__(self):
        return hash((self.peer, self.seeder_pk))

    def to_dict(self):
        return{'address': {'ip': self.peer.address[0],
                           'port': self.peer.address[1],
                           'public_key': hexlify(self.peer.public_key.key_to_bin()).decode()},
               'seeder_pk': hexlify(self.seeder_pk).decode(),
               'source': self.source}


class Swarm(object):

    def __init__(self, info_hash, hops, lookup_func, seeder_sk=None, max_ip_age=180,
                 min_dht_lookup_interval=300, max_dht_lookup_interval=120):
        self.info_hash = info_hash
        self.hops = hops
        self.lookup_func = lookup_func
        self.seeder_sk = seeder_sk
        self.max_ip_age = max_ip_age
        self.min_dht_lookup_interval = min_dht_lookup_interval
        self.max_dht_lookup_interval = max_dht_lookup_interval

        self.intro_points = []
        self.connections = {}
        self.last_lookup = 0
        self.last_dht_response = 0
        self.transfer_history = [0, 0]

        self.logger = logging.getLogger(self.__class__.__name__)

    @property
    def seeding(self):
        return bool(self.seeder_sk)

    @property
    def _active_circuits(self):
        return [c for c, _ in self.connections.values() if c.state == CIRCUIT_STATE_READY and c.e2e]

    def add_connection(self, rp_circuit, intro_point_used):
        if rp_circuit.circuit_id not in self.connections:
            self.connections[rp_circuit.circuit_id] = (rp_circuit, intro_point_used)

    def remove_connection(self, rp_circuit):
        removed = self.connections.pop(rp_circuit.circuit_id, None)
        if removed:
            circuit, _ = removed
            self.transfer_history[0] += circuit.bytes_up
            self.transfer_history[1] += circuit.bytes_down
        return bool(removed)

    def has_connection(self, seeder_pk):
        return seeder_pk in [ip.seeder_pk for _, ip in self.connections.values()]

    def get_num_connections(self):
        return len(self._active_circuits)

    def get_num_connections_incomplete(self):
        return len(self.connections) - self.get_num_connections()

    def add_intro_point(self, ip):
        old_ip = next((i for i in self.intro_points if i == ip), None)

        if old_ip:
            old_ip.last_seen = time.time()
        else:
            self.intro_points.append(ip)

        return old_ip or ip

    def remove_old_intro_points(self):
        # Cleanup old introduction points
        now = time.time()
        used_intro_points = [i for c, i in self.connections.values() if c.state == CIRCUIT_STATE_READY and c.e2e]
        self.intro_points = [i for i in self.intro_points
                             if i.last_seen + self.max_ip_age >= now or i in used_intro_points]

    def remove_intro_point(self, ip):
        try:
            self.intro_points.remove(ip)
        except ValueError:
            pass

    async def lookup(self, target=None):
        def on_success(ips):
            if any([ip for ip in ips if ip.source == PEER_SOURCE_DHT]):
                self.last_dht_response = time.time()
            return ips

        # Are we doing a manual lookup?
        if target:
            self.logger.info("Performing manual PEX lookup for swarm %s (target %s)",
                             hexlify(self.info_hash), target.peer)
            return on_success(await self.lookup_func(self.info_hash, target, self.hops))

        now = time.time()
        self.remove_old_intro_points()
        self.last_lookup = now
        if (now - self.last_dht_response) > self.min_dht_lookup_interval or \
           (not self.intro_points and (now - self.last_dht_response) > self.max_dht_lookup_interval):
            self.logger.info("Performing DHT lookup for swarm %s", hexlify(self.info_hash))
            return on_success(await self.lookup_func(self.info_hash, None, self.hops))
        elif self.intro_points:
            self.logger.info("Performing PEX lookup for swarm %s (targeting %d peer(s))",
                             hexlify(self.info_hash), len(self.intro_points))
            results = []
            tasks = [self.lookup_func(self.info_hash, ip, self.hops) for ip in self.intro_points]
            if tasks:
                results = [result for result in (await gather(*tasks, return_exceptions=True))
                           if not isinstance(result, (CancelledError, Exception))]
                results = sum(results, [])
            return on_success(results)
        self.logger.info("Skipping lookup for swarm %s", hexlify(self.info_hash))

    def get_num_seeders(self):
        seeder_pks = {ip.seeder_pk for ip in self.intro_points}
        for _, ip in self.connections.values():
            seeder_pks.add(ip.seeder_pk)
        return len(seeder_pks)

    def get_total_up(self):
        return sum([c.bytes_up for c in self._active_circuits]) + self.transfer_history[0]

    def get_total_down(self):
        return sum([c.bytes_down for c in self._active_circuits]) + self.transfer_history[1]



from random import randint, shuffle
from threading import Lock

from .structs import BitPairAttestation, BonehAttestation
from ..primitives.attestation import sha256_4_as_int, sha256_as_int, sha512_as_int
from ..primitives.boneh import decode, encode
from ..primitives.value import FP2Value

multithread_update_lock = Lock()


def generate_modular_additive_inverse(p, n):
    """
    Generate a group of size n which is its own modular additive inverse modulo p + 1.
    """
    R = [randint(1, p - 1) for _ in range(n - 1)]
    R.append(p - (sum(R) % (p + 1)) + 1)
    shuffle(R)
    return R


def attest(PK, value, bitspace):
    """
    Create an attestation for a public key's value lying within a certain bitspace.
    """
    A = list([int(c) for c in str(bin(value))[2:]])
    while len(A) < bitspace:
        A.insert(0, 0)
    R = generate_modular_additive_inverse(PK.p, bitspace)
    t_out_public = [encode(PK, a + b) for (a, b) in zip(A, R)]
    t_out_private = []
    for i in range(0, len(A) - 1, 2):
        t_out_private.append((i, encode(PK, PK.p - ((R[i] + R[i + 1]) % (PK.p + 1)) + 1)))
    # Shuffle:
    t_out_public = [(i, t_out_public[i], t_out_public[i + 1]) for i in range(0, len(t_out_public), 2)]
    shuffle(t_out_public)
    out_public = []
    out_private = []
    shuffle_map = {}
    for (i, v1, v2) in t_out_public:
        shuffle_map[i] = len(out_public)
        out_public.append(v1)
        out_public.append(v2)
    for (i, e) in t_out_private:
        out_private.append((shuffle_map[i], e))
    shuffle(out_private)
    # Formalize
    bitpairs = []
    for (i, e) in out_private:
        bitpairs.append(BitPairAttestation(out_public[i], out_public[i + 1], e))
    return BonehAttestation(PK, bitpairs)


def attest_sha512(PK, value):
    """
    Create an attestation for a value using a SHA512 hash.
    """
    return attest(PK, sha512_as_int(value), 512)


def binary_relativity_sha512(value):
    """
    Create the inter-bitpair relativity map of a value using the SHA512 hash.
    """
    return binary_relativity(sha512_as_int(value), 512)


def attest_sha256(PK, value):
    """
    Create an attestation for a value using a SHA256 hash.
    """
    return attest(PK, sha256_as_int(value), 256)


def binary_relativity_sha256(value):
    """
    Create the inter-bitpair relativity map of a value using the SHA256 hash.
    """
    return binary_relativity(sha256_as_int(value), 256)


def attest_sha256_4(PK, value):
    """
    Create an attestation for a value using a SHA256 4 byte hash.
    """
    return attest(PK, sha256_4_as_int(value), 32)


def binary_relativity_sha256_4(value):
    """
    Create the inter-bitpair relativity map of a value using the SHA256 4 byte hash.
    """
    return binary_relativity(sha256_4_as_int(value), 32)


def create_empty_relativity_map():
    """
    Construct a map of possible challenge responses.
    """
    return {0: 0, 1: 0, 2: 0, 3: 0}


def binary_relativity(value, bitspace):
    """
    Create the inter-bitpair relativity map of a value.
    """
    out = {0: 0, 1: 0, 2: 0}
    A = list([int(c) for c in str(bin(value))[2:]])
    while len(A) < bitspace:
        A.insert(0, 0)
    for i in range(0, bitspace - 1, 2):
        out[A[i] + A[i + 1]] += 1
    out[3] = 0
    return out


def binary_relativity_match(expected, value):
    """
    Get the matching percentage between relativity maps.
    Mismatches result in 0.0.
    """
    match = 1.0
    for k in expected:
        if expected[k] < value[k]:
            return 0.0
        if not expected[k] or not value[k]:
            continue
        match *= float(value[k]) / float(expected[k])
    return match


def binary_relativity_certainty(expected, value):
    """
    Give the chance of a current relativity map being the expected one.
    """
    cert = 1 - 0.5 ** (sum(value.values()))
    return binary_relativity_match(expected, value) * cert


def create_challenge(PK, bitpair):
    """
    Create a challenge for a bitpair attestation of a certain public key.
    """
    return bitpair.compress() * encode(PK, 0)


def create_honesty_check(PK, value):
    """
    Create a honesty check challenge.
    """
    return encode(PK, value)


def create_challenge_response_from_pair(SK, pair):
    """
    Respond to a bitpair challenge.
    """
    return create_challenge_response(SK, FP2Value(SK.p, pair[0], pair[1]))


def create_challenge_response(SK, challenge):
    """
    Respond to a bitpair challenge.
    """
    decoded = decode(SK, [0, 1, 2], challenge)
    return 3 if decoded is None else decoded


def process_challenge_response(relativity_map, response):
    """
    Process a challenge response in a relativity map.
    """
    multithread_update_lock.acquire()
    relativity_map[response] += 1
    multithread_update_lock.release()



"""
For most types of Community loading, you can use the ConfigBuilder.
These classes present an alternative way to load Community objects into IPv8, itself loaded into another object.
Use if you have a IPv8 manager (a "session") that needs fine-grained control over the loading of each community.
"""

import logging
import types

from .keyvault.crypto import default_eccrypto
from .peer import Peer


# pylint: disable=W0613


def name(str_name):
    """
    Specify a custom name for this launcher.

    For example:

     .. code-block :: Python

        @name('Community1')
        class A(CommunityLauncher):
           ...

    :param str_name: the new name to give this launcher.
    """
    def decorator(instance):
        def new_get_name(_):
            return str_name

        instance.get_name = new_get_name
        return instance
    return decorator


def after(*launcher_name):
    """
    Specify one or more Community classes which should be loaded before the CommunityLauncher is invoked.
    You may call this multiple times and/or with multiple arguments.

    For example:

     .. code-block :: Python

        @after('CommunityLauncher1', 'CommunityLauncher2')
        class A(CommunityLauncher):
           ...

        @after('CommunityLauncher1')
        @after('CommunityLauncher2')
        class A(CommunityLauncher):
           ...

    :param launcher_name: the launcher name(s) that need to be loaded beforehand.
    """
    def decorator(instance):
        old_not_before = instance.not_before

        def new_not_before(self):
            return old_not_before(self) + list(launcher_name)

        instance.not_before = new_not_before
        return instance
    return decorator


def precondition(str_condition):
    """
    Specify a string to be evaluated and interpreted as a condition for this CommunityLauncher to start.
    A ``session`` object is provided to pull a state from.
    You may call this multiple times.

    For example:

     .. code-block :: Python

        @precondition('session.some_condition')
        class A(CommunityLauncher):
           ...

        @precondition('session.some_condition')
        @precondition('not session.some_condition_method()')
        class A(CommunityLauncher):
           ...

    :param str_condition: the string to be evaluated as a launch condition.
    """
    def decorator(instance):
        old_should_launch = instance.should_launch

        def new_should_launch(self, session):
            return (old_should_launch(self, session)
                    and eval(str_condition, globals(), locals()))  # pylint: disable=W0123

        instance.should_launch = new_should_launch
        return instance
    return decorator


def _get_class(class_or_function):
    """
    Return the class by processing incoming argument either as a function or as
    a class.

    :param class_or_function: the class, or the function that represents this class.
    """
    return class_or_function() if isinstance(class_or_function, types.FunctionType) \
        else class_or_function


def overlay(str_module_or_class, str_definition=None):
    """
    Specify, as strings, a module and Community class object defined therein to lazy-load.
    Otherwise, give an actual Community class to load.

    For example:

     .. code-block :: Python

        @overlay('my_module.some_submodule', 'MyCommunityClass')
        class A(CommunityLauncher):
           ...

        from my_module.some_submodule import MyCommunityClass
        @overlay(MyCommunityClass)
        class A(CommunityLauncher):
           ...

        def my_community_class():
            from my_module.some_submodule import MyCommunityClass
            return MyCommunityClass

        @overlay(my_community_class)
        class A(CommunityLauncher):
           ...

    :param str_module_or_class: either the module to load or a Community class.
    :param str_definition: either the class definition to load or None if str_module_or_class is not a string.
    """
    def decorator(instance):
        if isinstance(str_module_or_class, str):
            if not hasattr(instance, "hiddenimports"):
                setattr(instance, "hiddenimports", set())
            instance.hiddenimports.add(str_module_or_class)

            def get_overlay_class(_):
                return getattr(__import__(str_module_or_class, fromlist=[str_definition]), str_definition)
        else:
            def get_overlay_class(_):
                return _get_class(str_module_or_class)

        instance.get_overlay_class = get_overlay_class
        return instance
    return decorator


def walk_strategy(str_module_or_class, str_definition=None, target_peers=20, kw_args=None):
    """
    Specify, as strings, a module and DiscoveryStrategy class object defined therein to lazy-load.
    Otherwise, give an actual DiscoveryStrategy class to load.

    For example:

     .. code-block :: Python

        @walk_strategy('my_module.some_submodule', 'MyStrategyClass')
        class A(CommunityLauncher):
           ...

        from my_module.some_submodule import MyStrategyClass
        @walk_strategy(MyStrategyClass)
        class A(CommunityLauncher):
           ...

        @walk_strategy('my_module.some_submodule', 'MyStrategyClass', target_peers=-1, kwargs={'a key': 'a value'})
        class A(CommunityLauncher):
           ...

       def my_strategy_class():
            from my_module.some_submodule import MyStrategyClass
            return MyStrategyClass

        @walk_strategy(my_strategy_class)
        class A(CommunityLauncher):
            ...

    :param str_module_or_class: either the module to load or a DiscoveryStrategy class.
    :param str_definition: either the class definition to load or None if str_module_or_class is not a string.
    :param target_peers: the target_peers for the strategy.
    :param kw_args: the keyword arguments to initialize the DiscoveryStrategy instance with.
    """
    def decorator(instance):
        old_get_walk_strategies = instance.get_walk_strategies

        if isinstance(str_module_or_class, str):
            if not hasattr(instance, "hiddenimports"):
                setattr(instance, "hiddenimports", set())
            instance.hiddenimports.add(str_module_or_class)
            strategy_class = getattr(__import__(str_module_or_class, fromlist=[str_definition]), str_definition)
        else:
            strategy_class = _get_class(str_module_or_class)

        def new_get_walk_strategies(self):
            return old_get_walk_strategies(self) + [(strategy_class, kw_args or {}, target_peers)]

        instance.get_walk_strategies = new_get_walk_strategies
        return instance
    return decorator


def bootstrapper(str_module_or_class, str_definition=None, kw_args=None):
    """
    Specify, as strings, a module and Bootstrapper class object defined therein to lazy-load.
    Otherwise, give an actual Bootstrapper class to load.

    For example:

     .. code-block :: Python

        @bootstrapper('my_module.some_submodule', 'MyBootstrapper')
        class A(CommunityLauncher):
           ...

        from my_module.some_submodule import MyBootstrapper
        @walk_strategy(MyBootstrapper)
        class A(CommunityLauncher):
           ...

        @walk_strategy('my_module.some_submodule', 'MyBootstrapper', kwargs={'a key': 'a value'})
        class A(CommunityLauncher):
           ...

       def my_bootstrapper_class():
            from my_module.some_submodule import MyBootstrapper
            return MyBootstrapper

        @walk_strategy(my_bootstrapper_class)
        class A(CommunityLauncher):
            ...

    :param str_module_or_class: either the module to load or a Bootstrapper class.
    :param str_definition: either the class definition to load or None if str_module_or_class is not a string.
    :param kw_args: the keyword arguments to initialize the Bootstrapper instance with.
    """
    def decorator(instance):
        old_get_bootstrappers = instance.get_bootstrappers

        if isinstance(str_module_or_class, str):
            if not hasattr(instance, "hiddenimports"):
                setattr(instance, "hiddenimports", set())
            instance.hiddenimports.add(str_module_or_class)
            bootstrapper_class = getattr(__import__(str_module_or_class, fromlist=[str_definition]), str_definition)
        else:
            bootstrapper_class = _get_class(str_module_or_class)

        def new_get_bootstrappers(self, session):
            return old_get_bootstrappers(self, session) + [(bootstrapper_class, kw_args or {})]

        instance.get_bootstrappers = new_get_bootstrappers
        return instance
    return decorator


def set_in_session(attribute_name):
    """
    Specify an attribute to set on the session, once the CommunityLauncher has finished initializing its Community.

    For example, the following sets the ``session.my_community`` to the loaded Community instance:

     .. code-block :: Python

        @set_in_session('my_community')
        class A(CommunityLauncher):
           ...

    :param attribute_name: the attribute name (string) to set on the session, once the Community is loaded.
    """
    def decorator(instance):
        old_finalize = instance.finalize

        def new_finalize(self, ipv8, session, community):
            out = old_finalize(self, ipv8, session, community)
            setattr(session, attribute_name, community)
            return out

        instance.finalize = new_finalize
        return instance
    return decorator


def kwargs(**kw_args):
    """
    Specify keyword arguments as evaluated strings, to initialize the Community with.
    A ``session`` object is provided to pull a state from.

    For example:

     .. code-block :: Python

        @kwargs('working_directory'='session.working_directory')
        class A(CommunityLauncher):
           ...

        @kwargs(a_key='"I am a string! :)"')
        class A(CommunityLauncher):
           ...

    :param kw_args: the mapping of keyword arguments to statements to be evaluated.
    """
    def decorator(instance):
        old_get_kwargs = instance.get_kwargs

        def new_get_kwargs(self, session):
            out = old_get_kwargs(self, session)
            for kwarg in kw_args:
                out[kwarg] = eval(kw_args[kwarg], globals(), locals())  # pylint: disable=W0123
            return out

        instance.get_kwargs = new_get_kwargs
        return instance

    return decorator


class CommunityLauncher:

    """
    Object in charge of preparing a Community for loading in IPv8.
    """

    def __init__(self):
        super()
        self.community_args = []
        self.community_kwargs = {}

    def get_name(self):
        """
        Get the launcher name, for pre-launch organisation.

        :rtype: str
        """
        return self.__class__.__name__

    def not_before(self):
        """
        Should not launch this before some other launcher has completed.

        :return: The list of launcher names to complete before this is launched
        """
        return []

    def should_launch(self, session):
        """
        Check whether this launcher should launch.

        For example:

            return session.config.get_tunnel_community_enabled()

        :type session: object
        :rtype: bool
        """
        return True

    def prepare(self, overlay_provider, session):
        """
        Perform setup tasks before the community is loaded.

        :type overlay_provider: ipv8.IPv8
        :type session: object
        """

    def finalize(self, ipv8, session, community):
        """
        Perform cleanup tasks after the community has been loaded.

        :type ipv8: ipv8.IPv8
        :type session: object
        :type community: IPv8 community
        """

    def get_args(self, session):
        """
        Get the args to load the community with.

        :rtype: tuple
        """
        return self.community_args

    def get_kwargs(self, session):
        """
        Get the kwargs to load the community with.

        :rtype: dict or None
        """
        ret = {}
        ret.update(self.community_kwargs)
        return ret

    def get_overlay_class(self):
        """
        Get the overlay class this launcher wants to load.

        This raises a RuntimeError if it was not overwritten at runtime, to appease Pylint.

        :rtype: ipv8.overlay.Overlay
        """
        raise RuntimeError("CommunityLaunchers should define an Overlay class to load!")

    def get_walk_strategies(self):
        """
        Get walk strategies for this class.
        It should be provided as a list of tuples with the class, kwargs and maximum number of peers.
        """
        return []

    def get_bootstrappers(self, session):
        """
        Get the bootstrappers for this class.
        It should be provided as a list of tuples with the class and kwargs.
        """
        return []

    def get_my_peer(self, ipv8, session):
        return Peer(default_eccrypto.generate_key("curve25519"))


class CommunityLoader:
    """
    Object in charge of loading communities into IPv8.
    """

    def __init__(self):
        self.community_launchers = {}
        self._logger = logging.getLogger(self.__class__.__name__)

    def get_launcher(self, name):
        return self.community_launchers[name][0]

    def has_launched(self, name):
        return self.community_launchers[name][1]

    def set_launcher(self, launcher):
        """
        Register a launcher to be launched by name.

        If a launcher for the same name already existed, it is overwritten.

        :type launcher: CommunityLauncher
        """
        assert isinstance(launcher, CommunityLauncher)

        if launcher.get_name() in self.community_launchers:
            self._logger.info("Overriding CommunityLauncher %s", launcher.get_name())
            if self.has_launched(launcher.get_name()):
                self._logger.error("Unable to replace launcher for %s, it was already launched", launcher.get_name())
                return

        self.community_launchers[launcher.get_name()] = (launcher, False)

    def del_launcher(self, launcher):
        """
        Unregister a launcher

        :type launcher: CommunityLauncher
        """
        assert isinstance(launcher, CommunityLauncher)

        if launcher.get_name() in self.community_launchers:
            del self.community_launchers[launcher.get_name()]

    def load(self, overlay_provider, session):
        """
        Load all of the communities specified by the registered launchers into IPv8.
        """
        remaining = [launcher for launcher, _ in self.community_launchers.values()]
        cycle = len(remaining) * len(remaining)
        while remaining and cycle >= 0:
            launcher = remaining.pop(0)
            cycle -= 1
            if launcher.should_launch(session):
                validated = True
                for dependency in launcher.not_before():
                    # If the dependency does not exist, don't wait for it
                    # If the dependency is never loaded, don't wait for it
                    if dependency in self.community_launchers and \
                            self.community_launchers[dependency][0].should_launch(session):
                        validated = validated and self.has_launched(dependency)
                if validated:
                    self._launch(launcher, overlay_provider, session)
                else:
                    remaining.append(launcher)
        if cycle < 0:
            launcher_names = [launcher.get_name() for launcher in remaining]
            raise RuntimeError("Cycle detected in CommunityLauncher not_before(): %s" % (str(launcher_names)))
        self._logger.info("Finished loading communities!")

    def _launch(self, launcher, ipv8, session):
        """
        This method should be overridden.
        """


class IPv8CommunityLoader(CommunityLoader):
    """
    Loader for IPv8 communities.
    """

    def _launch(self, launcher, ipv8, session):
        """
        Launch a launcher: register the overlay with IPv8.
        """
        # Prepare launcher
        launcher.prepare(ipv8, session)
        # Register community
        overlay_class = launcher.get_overlay_class()
        self._logger.info("Loading overlay %s", overlay_class)
        walk_strategies = launcher.get_walk_strategies()
        peer = launcher.get_my_peer(ipv8, session)
        args = launcher.get_args(session)
        kwargs = launcher.get_kwargs(session)
        overlay = overlay_class(peer, ipv8.endpoint, ipv8.network, *args, **kwargs)
        bootstrappers = launcher.get_bootstrappers(session)

        ipv8.overlays.append(overlay)
        for strategy_class, strategy_kwargs, target_peers in walk_strategies:
            ipv8.strategies.append((strategy_class(overlay, **strategy_kwargs), target_peers))
        for bootstrapper_class, bootstrapper_kwargs in bootstrappers:
            overlay.bootstrappers.append(bootstrapper_class(**bootstrapper_kwargs))

        # Cleanup
        launcher.finalize(ipv8, session, overlay)
        self.community_launchers[launcher.get_name()] = (launcher, True)



from functools import reduce
from struct import calcsize, pack, unpack_from

from ...messaging.anonymization.tunnel import (CIRCUIT_TYPE_RP_DOWNLOADER, CIRCUIT_TYPE_RP_SEEDER, EXIT_NODE,
                                               EXIT_NODE_SALT, ORIGINATOR, ORIGINATOR_SALT)
from ...messaging.anonymization.tunnelcrypto import CryptoException
from ...messaging.lazy_payload import VariablePayload, vp_compile

NO_CRYPTO_PACKETS = [2, 3]


@vp_compile
class ExtraIntroductionPayload(VariablePayload):
    names = ['flags']
    format_list = ['flags']


@vp_compile
class DataPayload(VariablePayload):
    msg_id = 1
    names = ['circuit_id', 'dest_address', 'org_address', 'data']
    format_list = ['I', 'address', 'address', 'raw']


@vp_compile
class CreatePayload(VariablePayload):
    msg_id = 2
    names = ['circuit_id', 'identifier', 'node_public_key', 'key']
    format_list = ['I', 'H', 'varlenH', 'varlenH']


@vp_compile
class CreatedPayload(VariablePayload):
    msg_id = 3
    names = ['circuit_id', 'identifier', 'key', 'auth', 'candidate_list_enc']
    format_list = ['I', 'H', 'varlenH', '32s', 'raw']


@vp_compile
class ExtendPayload(VariablePayload):
    msg_id = 4
    names = ['circuit_id', 'identifier', 'node_public_key', 'key', 'node_addr']
    format_list = ['I', 'H', 'varlenH', 'varlenH', 'ipv4']


@vp_compile
class ExtendedPayload(VariablePayload):
    msg_id = 5
    names = ['circuit_id', 'identifier', 'key', 'auth', 'candidate_list_enc']
    format_list = ['I', 'H', 'varlenH', '32s', 'raw']


@vp_compile
class PingPayload(VariablePayload):
    msg_id = 6
    names = ['circuit_id', 'identifier']
    format_list = ['I', 'H']


@vp_compile
class PongPayload(VariablePayload):
    msg_id = 7
    names = ['circuit_id', 'identifier']
    format_list = ['I', 'H']


@vp_compile
class DestroyPayload(VariablePayload):
    msg_id = 8
    names = ['circuit_id', 'reason']
    format_list = ['I', 'H']


@vp_compile
class EstablishIntroPayload(VariablePayload):
    msg_id = 9
    names = ['circuit_id', 'identifier', 'info_hash', 'public_key']
    format_list = ['I', 'H', '20s', 'varlenH']


@vp_compile
class IntroEstablishedPayload(VariablePayload):
    msg_id = 10
    names = ['circuit_id', 'identifier']
    format_list = ['I', 'H']


@vp_compile
class EstablishRendezvousPayload(VariablePayload):
    msg_id = 11
    names = ['circuit_id', 'identifier', 'cookie']
    format_list = ['I', 'H', '20s']


@vp_compile
class RendezvousEstablishedPayload(VariablePayload):
    msg_id = 12
    names = ['circuit_id', 'identifier', 'rendezvous_point_addr']
    format_list = ['I', 'H', 'ipv4']


@vp_compile
class CreateE2EPayload(VariablePayload):
    msg_id = 13
    names = ['identifier', 'info_hash', 'node_public_key', 'key']
    format_list = ['H', '20s', 'varlenH', 'varlenH']


@vp_compile
class CreatedE2EPayload(VariablePayload):
    msg_id = 14
    names = ['identifier', 'key', 'auth', 'rp_info_enc']
    format_list = ['H', 'varlenH', '32s', 'raw']


@vp_compile
class LinkE2EPayload(VariablePayload):
    msg_id = 15
    names = ['circuit_id', 'identifier', 'cookie']
    format_list = ['I', 'H', '20s']


@vp_compile
class LinkedE2EPayload(VariablePayload):
    msg_id = 16
    names = ['circuit_id', 'identifier']
    format_list = ['I', 'H']


@vp_compile
class PeersRequestPayload(VariablePayload):
    msg_id = 17
    names = ['circuit_id', 'identifier', 'info_hash']
    format_list = ['I', 'H', '20s']


@vp_compile
class IntroductionInfo(VariablePayload):
    names = ['address', 'key', 'seeder_pk', 'source']
    format_list = ['ipv4', 'varlenH', 'varlenH', 'B']


@vp_compile
class PeersResponsePayload(VariablePayload):
    msg_id = 18
    names = ['circuit_id', 'identifier', 'info_hash', 'peers']
    format_list = ['I', 'H', '20s', [IntroductionInfo]]


@vp_compile
class RendezvousInfo(VariablePayload):
    names = ['address', 'key', 'cookie']
    format_list = ['ipv4', 'varlenH', '20s']


@vp_compile
class TestRequestPayload(VariablePayload):
    msg_id = 19
    names = ['circuit_id', 'identifier', 'response_size', 'data']
    format_list = ['I', 'H', 'H', 'raw']


@vp_compile
class TestResponsePayload(VariablePayload):
    msg_id = 20
    names = ['circuit_id', 'identifier', 'data']
    format_list = ['I', 'H', 'raw']


class Flags:

    def __init__(self, fmt='>H'):
        self.format = fmt
        self.size = calcsize(fmt)

    def pack(self, data):
        return pack(self.format, reduce(lambda a, b: a | b, data, 0))

    def unpack(self, data, offset, unpack_list):
        number, = unpack_from(self.format, data, offset)
        unpack_list.append(list(filter(None, [number & (2 ** i) for i in range(self.size * 8)])))
        return self.size


class CellPayload:
    msg_id = 0

    def __init__(self, circuit_id, message, plaintext=False, relay_early=False):
        self.circuit_id = circuit_id
        self.message = message
        self.plaintext = plaintext
        self.relay_early = relay_early

    def encrypt(self, crypto, circuit=None, relay_session_keys=None):
        if self.plaintext:
            return

        if circuit:
            if circuit.hs_session_keys and circuit.ctype in [CIRCUIT_TYPE_RP_SEEDER, CIRCUIT_TYPE_RP_DOWNLOADER]:
                direction = int(circuit.ctype == CIRCUIT_TYPE_RP_SEEDER)
                self.message = crypto.encrypt_str(self.message,
                                                  *crypto.get_session_keys(circuit.hs_session_keys, direction))

            for hop in reversed(circuit.hops):
                self.message = crypto.encrypt_str(self.message, *crypto.get_session_keys(hop.session_keys, EXIT_NODE))

        elif relay_session_keys:
            self.message = crypto.encrypt_str(self.message, *crypto.get_session_keys(relay_session_keys, ORIGINATOR))

        else:
            raise CryptoException("Error encrypting message for unknown circuit %d" % self.circuit_id)

    def decrypt(self, crypto, circuit=None, relay_session_keys=None):
        if self.plaintext:
            return

        if circuit:
            if not circuit.hops:
                raise CryptoException("Error decrypting message for 0-hop circuit %d" % self.circuit_id)

            # Remove all the encryption layers
            for layer, hop in enumerate(circuit.hops):
                try:
                    self.message = crypto.decrypt_str(self.message,
                                                      hop.session_keys[ORIGINATOR],
                                                      hop.session_keys[ORIGINATOR_SALT])
                except ValueError as e:
                    raise CryptoException("Got exception %r when trying to remove encryption layer %s "
                                          "for message: %r received for circuit_id: %s, circuit_hops: %r" %
                                          (e, layer, self.message, self.circuit_id, circuit.hops)) from e

            if circuit.hs_session_keys and circuit.ctype in [CIRCUIT_TYPE_RP_SEEDER, CIRCUIT_TYPE_RP_DOWNLOADER]:
                direction = int(circuit.ctype == CIRCUIT_TYPE_RP_DOWNLOADER)
                direction_salt = direction + 2
                self.message = crypto.decrypt_str(self.message,
                                                  circuit.hs_session_keys[direction],
                                                  circuit.hs_session_keys[direction_salt])

        elif relay_session_keys:
            try:
                self.message = crypto.decrypt_str(self.message,
                                                  relay_session_keys[EXIT_NODE],
                                                  relay_session_keys[EXIT_NODE_SALT])
            except ValueError as e:
                raise CryptoException("Got exception %r when trying to decrypt relay message: "
                                      "cell received for circuit_id: %s" % (e, self.circuit_id)) from e

        else:
            raise CryptoException("Error decrypting message for unknown circuit %d" % self.circuit_id)

    def unwrap(self, prefix):
        return b''.join([prefix,
                         self.message[0:1],
                         pack('!I', self.circuit_id),
                         self.message[1:]])

    def to_bin(self, prefix):
        return b''.join([prefix,
                         bytes([self.msg_id]),
                         pack('!I??', self.circuit_id, self.plaintext, self.relay_early) + self.message])

    @classmethod
    def from_bin(cls, packet):
        circuit_id, plaintext, relay_early = unpack_from('!I??', packet, 23)
        return cls(circuit_id, packet[29:], plaintext, relay_early)
